@@ -7,7 +7,7 @@ SET pg_hint_plan.enable_hint TO on;
 CREATE EXTENSION file_fdw;
 CREATE SERVER file_server FOREIGN DATA WRAPPER file_fdw;
 CREATE USER MAPPING FOR PUBLIC SERVER file_server;
-CREATE FOREIGN TABLE ft1 (id int, val int) SERVER file_server OPTIONS (format 'csv', filename '/home/horiguti/work/pg_hint_plan/data/data.csv');
+CREATE FOREIGN TABLE ft1 (id int, val int) SERVER file_server OPTIONS (format 'csv', filename '/home/horiguti/work/pg_hint_plan/pg_hint_plan/data/data.csv');
 
 -- foreign table test
 SELECT * FROM ft1;
@@ -43,9 +43,11 @@ EXPLAIN (COSTS false) SELECT * FROM s1.t1, ft1 ft_1, ft1 ft_2 WHERE t1.c1 = ft_1
 ----
 
 -- No. R-1-6-6
-\o results/ut-fdw.tmpout
+\o results/R_1-6-6.out.log
 /*+Rows(ft_1 ft_2 #1)Leading(ft_1 ft_2 t1)*/
 EXPLAIN SELECT * FROM s1.t1, ft1 ft_1, ft1 ft_2 WHERE t1.c1 = ft_1.id AND t1.c1 = ft_2.id;
 \o
-\! sql/maskout.sh results/ut-fdw.tmpout
-\! rm results/ut-fdw.tmpout
+\! sed 's/cost=[\.0-9]*/cost=xxx/' results/R_1-6-6.out.log > results/R_1-6-6.out_2.log
+\! sed 's/Foreign File: .*/Foreign File: @pg_hint_plan@\/data\/data.csv/' results/R_1-6-6.out_2.log > results/R_1-6-6.out
+\! diff expected/R_1-6-6.out results/R_1-6-6.out
+
